from typing import Optional

from monopoly.processor import StatementProcessor


class BankBase(StatementProcessor):
    """Helper class to handle initialization of common variables
    that are shared between bank classes"""

    def __init__(self, file_path: str, password: Optional[str] = None):
        if password:
            # overwrite pydantic password with user-supplied password
            self.pdf_config.password = password

        # optional config
        brute_force_config = getattr(self, "brute_force_config", None)
        pdf_config = getattr(self, "pdf_config", None)

        super().__init__(
            statement_config=self.statement_config,
            brute_force_config=brute_force_config,
            pdf_config=pdf_config,
            file_path=file_path,
        )



import os
from enum import StrEnum, auto

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# pylint: disable=unused-argument,no-self-argument
class AutoEnum(StrEnum):
    """Generates lower case values for enums
    e.g. CITIBANK -> citibank
    """

    def _generate_next_value_(name: str, *args):  # type: ignore
        return name.lower()


class AccountType(AutoEnum):
    CREDIT = auto()


class BankNames(AutoEnum):
    CITIBANK = auto()
    DBS = auto()
    HSBC = auto()
    OCBC = auto()
    STANDARD_CHARTERED = auto()
    EXAMPLE = auto()


class StatementFields(AutoEnum):
    TRANSACTION_DATE = auto()
    DESCRIPTION = auto()
    AMOUNT = auto()


class SharedPatterns(StrEnum):
    AMOUNT = r"(?P<amount>[\d.,]+)$"
    DESCRIPTION = r"(?P<description>.*?)\s+"


class TransactionPatterns(StrEnum):
    DBS = (
        r"^(?P<transaction_date>\d{2}\s[A-Z]{3})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT
    )
    CITIBANK = (
        r"^(?P<transaction_date>\b\d{2}\s\w{3}\b)\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT
    )
    HSBC = (
        r"^(?P<posting_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
        r"(?P<transaction_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT
    )
    OCBC = (
        r"^(?P<transaction_date>\d+/\d+)\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT
    )
    STANDARD_CHARTERED = (
        r"^(?P<posting_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
        r"(?P<transaction_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?P<transaction_ref>Transaction\sRef\s\d+)\s+"
        + SharedPatterns.AMOUNT
    )



import logging
from datetime import datetime
from typing import Optional

from pandas import DataFrame

from monopoly.config import BruteForceConfig, StatementConfig
from monopoly.constants import StatementFields
from monopoly.pdf import PdfConfig, PdfParser
from monopoly.statement import Statement
from monopoly.storage import write_to_csv

logger = logging.getLogger(__name__)


class StatementProcessor(PdfParser):
    def __init__(
        self, statement_config, file_path, pdf_config=None, brute_force_config=None
    ):
        self.statement_config: StatementConfig = statement_config
        self.pdf_config: PdfConfig = pdf_config
        self.brute_force_config: BruteForceConfig = brute_force_config

        # provide access to get_pages()
        super().__init__(file_path, brute_force_config, pdf_config)

    def extract(self) -> Statement:
        pages = self.get_pages(self.brute_force_config)
        statement = Statement(pages, self.statement_config)

        if not statement.transactions:
            raise ValueError("No transactions found - statement extraction failed")

        if not statement.statement_date:
            raise ValueError("No statement date found")

        return statement

    def transform(self, statement: Statement) -> DataFrame:
        logger.info("Running transformation functions on DataFrame")
        df = self._transform_date_to_iso(statement.df, statement.statement_date)

        return df

    def _transform_date_to_iso(
        self, df: DataFrame, statement_date: datetime
    ) -> DataFrame:
        logger.info("Transforming dates to ISO 8601")
        df[StatementFields.TRANSACTION_DATE] = df.apply(
            self._convert_date, statement_date=statement_date, axis=1
        )
        return df

    def parse_date(self, date_str):
        date_format = self.statement_config.transaction_date_format
        parsed_date = datetime.strptime(date_str, date_format)
        return parsed_date.day, parsed_date.month

    def _convert_date(self, row, statement_date: datetime):
        row_day, row_month = self.parse_date(row[StatementFields.TRANSACTION_DATE])

        # Deal with mixed years from Jan/Dec
        if statement_date.month == 1 and row_month == 12:
            row_year = statement_date.year - 1
        else:
            row_year = statement_date.year

        return f"{row_year}-{row_month:02d}-{row_day:02d}"

    def load(
        self, df: DataFrame, statement: Statement, csv_file_path: Optional[str] = None
    ):
        csv_file_path = write_to_csv(
            df=df, csv_file_path=csv_file_path, statement=statement
        )

        return csv_file_path



import logging

from monopoly.config import PdfConfig, StatementConfig, settings
from monopoly.constants import AccountType, BankNames, TransactionPatterns

from ..base import BankBase

logger = logging.getLogger(__name__)


class Ocbc(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.OCBC,
        account_type=AccountType.CREDIT,
        transaction_pattern=TransactionPatterns.OCBC,
        transaction_date_format="%d/%m",
        statement_date_pattern=r"\d{2}\-\d{2}\-\d{4}",
        statement_date_format=r"%d-%m-%Y",
    )

    pdf_config = PdfConfig(
        password=settings.ocbc_pdf_password,
        page_range=(0, -2),
        page_bbox=(0, 0, 560, 750),
    )



import logging
import subprocess
from dataclasses import dataclass
from io import BytesIO
from typing import Optional

import fitz
import pdftotext
from pdf2john import PdfHashExtractor

from monopoly.config import BruteForceConfig, PdfConfig

logger = logging.getLogger(__name__)


@dataclass
class PdfPage:
    raw_text: str

    @property
    def lines(self) -> list[str]:
        return list(filter(None, self.raw_text.split("\n")))


class PdfParser:
    def __init__(
        self,
        file_path: str,
        brute_force_config: Optional[BruteForceConfig] = None,
        pdf_config: Optional[PdfConfig] = None,
    ):
        """Class responsible for parsing PDFs and returning raw text

        The page_range variable determines which pages are extracted.
        All pages are extracted by default.
        """
        self.file_path = file_path

        if pdf_config is None:
            pdf_config = PdfConfig()

        self.password = pdf_config.password
        self.page_range = slice(*pdf_config.page_range)
        self.page_bbox = pdf_config.page_bbox
        self.brute_force_config = brute_force_config

    def open(self, brute_force_config: Optional[BruteForceConfig] = None):
        """
        Opens and decrypts a PDF document
        """
        logger.info("Opening pdf from path %s", self.file_path)
        document = fitz.Document(self.file_path)

        if not document.is_encrypted:
            return document

        if self.password:
            document.authenticate(self.password)

            if document.is_encrypted:
                raise ValueError("Wrong password - unable to open document")

            return document

        # This attempts to unlock statements based on a common password,
        # followed by the last few digits of a card
        if brute_force_config:
            logger.info("Unlocking PDF using a string prefix with mask")
            password = self.unlock_pdf(
                pdf_file_path=self.file_path,
                static_string=brute_force_config.static_string,
                mask=brute_force_config.mask,
            )

            document.authenticate(password)

            if not document.is_encrypted:
                logger.info("Successfully authenticated with password")
                return document

            # If no successful authentication, raise an error
            raise ValueError(
                "Unable to unlock PDF password using static string and mask"
            )

        raise RuntimeError("Failed to open document")

    def get_pages(self, brute_force_config=None) -> list[PdfPage]:
        logger.info("Extracting text from PDF")
        document: fitz.Document = self.get_document(brute_force_config)

        num_pages = list(range(document.page_count))
        document.select(num_pages[self.page_range])

        for page in document:
            if self.page_bbox:
                logger.debug("Cropping page")
                page.set_cropbox(self.page_bbox)

            logger.debug("Removing vertical text")
            page = self._remove_vertical_text(page)

        pdf_byte_stream = BytesIO(document.tobytes())
        pdf = pdftotext.PDF(pdf_byte_stream, physical=True)

        return [PdfPage(page) for page in pdf]

    def get_document(self, brute_force_config=None) -> list[fitz.Page]:
        document: fitz.Document = self.open(brute_force_config)
        return document

    @staticmethod
    def _remove_vertical_text(page: fitz.Page):
        """Helper function to remove vertical text, based on writing direction (wdir).

        This helps avoid situations where the PDF is oddly parsed, due to vertical text
        inside the PDF.

        An example of vertical text breaking a transaction:
        ```
        'HEALTHY HARVEST CAFÉ SINGAPORE SG',
        'Co Reg No: 123456',
        '10 NOV 9.80',
        ```

        Note:
            The 'dir' key represents the tuple (cosine, sine) for the angle.
            If line["dir"] != (1, 0), the text of its spans is rotated.

        """
        for block in page.get_text("dict", flags=fitz.TEXTFLAGS_TEXT)["blocks"]:
            for line in block["lines"]:
                writing_direction = line["dir"]
                if writing_direction != (1, 0):
                    page.add_redact_annot(line["bbox"])
        page.apply_redactions(images=fitz.PDF_REDACT_IMAGE_NONE)
        return page

    @staticmethod
    def unlock_pdf(
        pdf_file_path: str, static_string: Optional[str], mask: Optional[str]
    ):
        hash_extractor = PdfHashExtractor(pdf_file_path)
        pdf_hash = hash_extractor.parse()

        hash_path = ".hash"
        with open(hash_path, "w", encoding="utf-8") as file:
            file.write(pdf_hash)

        mask_command = [
            f"john --format=PDF --mask={static_string}{mask} {hash_path} --pot=.pot"
        ]
        process = subprocess.run(mask_command, shell=True, check=False)

        if not process.returncode == 0:
            raise ValueError(f"Return code is not 0: {process}")

        show_command = ["john", "--show", hash_path, "--pot=.pot"]
        output = subprocess.run(
            show_command, capture_output=True, text=True, check=False
        )

        if not output.returncode == 0:
            raise ValueError(f"Return code is not 0: {output}")

        if "1 password hash cracked, 0 left" not in output.stdout:
            raise ValueError(f"PDF was not unlocked: {output}")

        password = output.stdout.split("\n")[0].split(":")[-1]

        return password



import logging
import re
from datetime import datetime
from functools import cached_property

from pandas import DataFrame
from pydantic import field_validator
from pydantic.dataclasses import dataclass

from monopoly.config import StatementConfig, arbitrary_config
from monopoly.constants import StatementFields
from monopoly.pdf import PdfPage

logger = logging.getLogger(__name__)


@dataclass
class Transaction:
    transaction_date: str
    description: str
    amount: float

    @field_validator("description", mode="after")
    def remove_extra_whitespace(cls, value: str) -> str:
        return " ".join(value.split())

    @field_validator("amount", mode="before")
    def adjust_number_format(cls, value: str) -> str:
        if isinstance(value, str):
            return value.replace(",", "")
        return value


@dataclass(config=arbitrary_config)
class Statement:
    pages: list[PdfPage]
    columns = [enum.value for enum in StatementFields]
    config: StatementConfig

    @cached_property
    def transactions(self) -> list[Transaction]:
        transactions = []
        for page in self.pages:
            lines = self.process_lines(page)
            for i, line in enumerate(lines):
                transaction = self._process_line(line, lines, idx=i)
                if transaction:
                    transactions.append(transaction)

        return transactions

    @staticmethod
    def process_lines(page: PdfPage) -> list:
        return [line.lstrip() for line in page.lines]

    def _process_line(
        self, line: str, lines: list[str], idx: int
    ) -> Transaction | None:
        if match := re.search(self.config.transaction_pattern, line):
            transaction = Transaction(**match.groupdict())  # type: ignore

            if self.config.multiline_transactions and idx < len(lines) - 1:
                next_line = lines[idx + 1]
                if not re.search(self.config.transaction_pattern, next_line):
                    transaction.description += " " + next_line
                    transaction = Transaction(**vars(transaction))

            return transaction
        return None

    @cached_property
    def statement_date(self):
        logger.info("Extracting statement date")
        first_page = self.pages[0]
        for line in first_page.lines:
            if match := re.findall(self.config.statement_date_pattern, line):
                statement_date = match[0]
                logger.debug("Statement date found")
                return datetime.strptime(
                    statement_date, self.config.statement_date_format
                )
        return None

    @cached_property
    def df(self) -> DataFrame:
        return DataFrame(self.transactions, columns=self.columns)



import logging

from monopoly.config import PdfConfig, StatementConfig
from monopoly.constants import AccountType, BankNames, TransactionPatterns

from ..base import BankBase

logger = logging.getLogger(__name__)


class Dbs(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.DBS,
        account_type=AccountType.CREDIT,
        transaction_pattern=TransactionPatterns.DBS,
        transaction_date_format="%d %b",
        statement_date_pattern=r"(\d{2}\s[A-Za-z]{3}\s\d{4})",
        statement_date_format=r"%d %b %Y",
    )

    pdf_config = PdfConfig(page_range=(0, -1))



import logging

from monopoly.config import PdfConfig, StatementConfig, settings
from monopoly.constants import AccountType, BankNames, TransactionPatterns

from ..base import BankBase

logger = logging.getLogger(__name__)


class StandardChartered(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.STANDARD_CHARTERED,
        account_type=AccountType.CREDIT,
        transaction_pattern=TransactionPatterns.STANDARD_CHARTERED,
        transaction_date_format="%d %b",
        statement_date_pattern=r"\d{2}\s\w+\s\d{4}",
        statement_date_format=r"%d %B %Y",
    )

    pdf_config = PdfConfig(
        password=settings.standard_chartered_pdf_password,
        page_range=(0, -1),
        page_bbox=(0, 0, 580, 820),
    )



import logging

from monopoly.config import PdfConfig, StatementConfig, settings
from monopoly.constants import AccountType, BankNames, TransactionPatterns

from ..base import BankBase

logger = logging.getLogger(__name__)


class Citibank(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.CITIBANK,
        account_type=AccountType.CREDIT,
        transaction_pattern=TransactionPatterns.CITIBANK,
        transaction_date_format="%d %b",
        statement_date_pattern=r"Statement\sDate\s+(.*)",
        statement_date_format=r"%B %d, %Y",
    )

    pdf_config = PdfConfig(
        password=settings.citibank_pdf_password,
        page_bbox=(20, 0, 595, 840),
        page_range=(0, -3),
    )



from .citibank import Citibank
from .dbs import Dbs
from .hsbc import Hsbc
from .ocbc import Ocbc
from .standard_chartered import StandardChartered

__all__ = ["Citibank", "Dbs", "Hsbc", "Ocbc", "StandardChartered"]



import logging

from monopoly.config import BruteForceConfig, PdfConfig, StatementConfig, settings
from monopoly.constants import AccountType, BankNames, TransactionPatterns

from ..base import BankBase

logger = logging.getLogger(__name__)


class Hsbc(BankBase):
    statement_config = StatementConfig(
        bank_name=BankNames.HSBC,
        account_type=AccountType.CREDIT,
        transaction_pattern=TransactionPatterns.HSBC,
        transaction_date_format="%d %b",
        statement_date_pattern=r"(\d{2}\s[A-Z]{3}\s\d{4})\s.*$",
        multiline_transactions=True,
        statement_date_format=r"%d %b %Y",
    )

    pdf_config = PdfConfig(
        page_range=(0, -1),
        page_bbox=(0, 0, 379, 842),
    )

    brute_force_config = BruteForceConfig(
        static_string=settings.hsbc_pdf_password_prefix, mask="?d?d?d?d?d?d"
    )
